%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% This is a rough, preliminary implementation of the frequency
% estimation algorithm presented in the paper "Frequency estimation using time
% domain methods based on robust differential operators", ICSP, Beijing,
% 2010, available at
% http://www.cse.unsw.edu.au/~ignjat/diff/frequencyEstimation.pdf
% An older, but more detailed Mathematica tutorial which also does
% detection of fragments can be found at
% http://www.cse.unsw.edu.au/~ignjat/diff/
% as well as papers on chromatic derivatives and expansions
% Qestions are welcome; you can email me at:
% ignjat@cse.unsw.edu.au
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

close all
clear all
clc

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% PARAMETERS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

numsin = 2;         % numsin <= 4;  number of sinusoidal components present in the signal

test_case = 4; % 1,2 or 3 - tests frequency separation, low amplitude sensitivity etc, see below

damped = 0;         % choosing if the sinusoids are damped or not

patching = 1;       %
                    % if patching == 1 the signal is patched at both ends;
% see   A. Ignjatovic: Signal interpolation using
% numerically robust differential operators, 14th WSEAS
% CSCC Multi-conference, July 22-25, 2010, Corfu
% Island, Greece, available at the above website.

win_type = 0;       % if patching ==0 selects a window type:
% 0 - rectangular;
% 1 - CD32 (a window obtained using chromatic expansion,
%     good for accurate differentiation)
% 2 - kaiser(3);
% 3 - hamming



%%%%%%%%%%%%%%%%%%%%%%%
% specifying the signal
%%%%%%%%%%%%%%%%%%%%%%%


if numsin == 1  % number of sinusoidal components in the signal
    if  test_case == 1
        SNR = 10;           % signal to noise ratio in db
        N = 32;            % number of signal samples, possibly minus 1, since we prefer odd number of samples
        f = 1;
    elseif  test_case == 2
        SNR = 20;           % signal to noise ratio in db
        N = 128;            % number of signal samples, possibly minus 1, since we prefer odd number of samples
    elseif  test_case == 3
        SNR = 500;           % signal to noise ratio in db
        N = 32;            % number of signal samples, possibly minus 1, since we prefer odd number of samples
        f = .05;    % .95*pi/2*rand(1,numsin);     % nomalised frequencies for the signal components
        
     elseif  test_case == 4
        SNR = 500;           % signal to noise ratio in db
        N = 32;            % number of signal samples, possibly minus 1, since we prefer odd number of samples
        f = 1.5;    % .95*pi/2*rand(1,numsin);     % nomalised frequencies for the signal components    
    end
   
    
    ns = 10^(-SNR/20);   % RMS of the noise
   
    phi = pi/2;                              % 2*pi*(rand(1,numsin)-0.5);   % generate random phase shifts between -pi and pi
    A = ones(1,numsin);                % signal amplitude
    
elseif numsin == 2        % number of sinusoidal components in the signal
    
    if test_case == 1
        N=128;
        SNR = 25;
        ns = 10^(-SNR/20);   % RMS of the noise
        f = [.4 1.4];     % .95*pi/2*rand(1,numsin);     % nomalised frequencies for the signal components
        phi = [0 1];      % 2*pi*(rand(1,numsin)-0.5);   % generate random phase shifts between -pi and pi
        A = [1 1];     % ones(1,numsin);              % signal amplitude
        
    elseif test_case == 2
        N=128;
        SNR = 35;
        ns = 10^(-SNR/20);   % RMS of the noise
        f = [.33 1.4];     % .95*pi/2*rand(1,numsin);     % nomalised frequencies for the signal components
        phi = 2*pi*(rand(1,numsin)-0.5);   % generate random phase shifts between -pi and pi
        A = [4*ns 1];     % ones(1,numsin);              % signal amplitude
        
        
    elseif test_case == 3
        N=128;
        SNR = 40;
        ns = 10^(-SNR/20);   % RMS of the noise
        f = [.9 1.4];     % .95*pi/2*rand(1,numsin);     % nomalised frequencies for the signal components
        phi = [0 1];      % 2*pi*(rand(1,numsin)-0.5);   % generate random phase shifts between -pi and pi
        A = [2*ns 1];     % ones(1,numsin);              % signal amplitude
      
    elseif test_case == 4
        N=128;
        SNR = 40;
        ns = 10^(-SNR/20);   % RMS of the noise
        f = [.9 1.4];     % .95*pi/2*rand(1,numsin);     % nomalised frequencies for the signal components
        phi = [0 1];      % 2*pi*(rand(1,numsin)-0.5);   % generate random phase shifts between -pi and pi
        A = [ns/sqrt(2) 1];     % ones(1,numsin);              % signal amplitude
        
        
    elseif test_case == 5
        N = 128; 
        SNR = 25;
        ns = 10^(-SNR/20);   % RMS of the noise
        f = [1.4 1.45];     % .95*pi/2*rand(1,numsin);     % nomalised frequencies for the signal components
        phi = [0 1];      % 2*pi*(rand(1,numsin)-0.5);   % generate random phase shifts between -pi and pi
        A = [.8 1];     % ones(1,numsin);              % signal amplitude
        
    end
    
elseif numsin == 3 && test_case == 1
    
    N = 64;
    SNR = 45;
    ns = 10^(-SNR/20);   % RMS of the noise
    %     f(1) = .005+rand(1,1)*pi/6;
    %     f(2) = f(1)+.15+rand(1,1)*pi/6;
    %     f(3) = f(2)+.15+rand(1,1)*(pi/2-f(2)-.15);
    f = [1.1 1.3 1.5];  %.95*pi/6*rand(1,numsin)+[pi/6 2*pi/6 3*pi/6];   %;    % nomalised frequencies for the signal components
    phi = [0 1 2];    % 2*pi*(rand(1,numsin)-0.5);   % generate random phase shifts between -pi and pi
    A = [.25 1 .2];      % ones(1,numsin);              % signal amplitude
    
 elseif numsin == 3 && test_case == 2  
    N = 32;
    SNR = 45;
    ns = 10^(-SNR/20);   % RMS of the noise
    %     f(1) = .005+rand(1,1)*pi/6;
    %     f(2) = f(1)+.15+rand(1,1)*pi/6;
    %     f(3) = f(2)+.15+rand(1,1)*(pi/2-f(2)-.15);
    
    f = [.8 1.1 1.2];  %.95*pi/6*rand(1,numsin)+[pi/6 2*pi/6 3*pi/6];   %;    % nomalised frequencies for the signal components
    phi = [0 1 2];    % 2*pi*(rand(1,numsin)-0.5);   % generate random phase shifts between -pi and pi
    A = [.25 1 .6];      % ones(1,numsin);              % signal amplitude
    
elseif numsin == 4 && test_case == 1  
    N = 64;
    SNR = 35;
    ns = 10^(-SNR/20);   % RMS of the noise
    f = [.2 .6 1.1 1.5];     % .95*pi/2*rand(1,numsin);     % nomalised frequencies for the signal components
    phi = [0 1 2 3];    % 2*pi*(rand(1,numsin)-0.5);   % generate random phase shifts between -pi and pi
    A = [1 1 1 1];    % ones(1,numsin);              % signal amplitude
elseif numsin == 4 && test_case == 2  
    N = 32;
    SNR = 40;
    ns = 10^(-SNR/20);   % RMS of the noise
    f = [.5 .65 1.3 1.45];     % .95*pi/2*rand(1,numsin);     % nomalised frequencies for the signal components
    phi = [0 1 2 3];    % 2*pi*(rand(1,numsin)-0.5);   % generate random phase shifts between -pi and pi
    A = [1 .5 .2 1];    % ones(1,numsin);              % signal amplitude
end

 f = sort(f);  % sorting for easy comparison with the estimates

 nsi = sqrt(1/2)*ns;  % RMS of the in-band component of the noise (signal is twice oversampled!)
 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%
nn = floor(N/2);
NN = 2*nn+1;             % making sure the number of samples is odd - convenient for work

INT = max(2*floor(N/16)+1,4);  % interval over which the correlation of chromatic
% derivatives is calculated; see below, to be experimented
% with; must be an odd integer

interval = zeros(1,3);
interval(1) = INT;      % lenghts of the intervals over which we calculate the correlation of k-th CDs,
interval(2) = 2*INT-1;  % to be experimented with; factors 2 and 4 are quite arbitary
interval(3) = 4*INT-1;
interval(4) = 4*INT-1;

if damped == 1
    damp = -2/NN*log(1/2);  % maximal absolute value of the damping factor
else
    damp = 0;
end

dmp = (2*rand(1,numsin)-1)*damp;  % choosing the damp factors randomly between -damp and +damp




for kk = numsin:4
    interval(kk) = interval(numsin);
end
% aver_int =max(INT,4);         % length of the interval over which to average eigenvecrors to estimate the 'rumble'
% has no impact on accuracy in the present
% implementation

K = nn-4; %nn-2*INT-2; % half the size of the interval over which to find the median of the eigenvectors


% Loading various filters:
%%%%%%%%%%%%%%%%%%%%%%%%%%%%

CD = load_CDfilters;        % A chromatic derivative filterbank: 129 taps; bandwidth 1.1*pi; pass
% 1.05*pi; transition .1*pi, for differentiation of noisy signals

lcd = (size(CD,2)-1)/2;     % getting half the number of taps of CD filters


if patching == 1
    
    CDClean = load_CDClean; % Loading a chromatic derivative filterbank: 129 taps; bandwidth pi; pass
    % pi; transition .2*pi, analysis filters for differentiation of clean signals
    
    [LPatch, RPatch]= load_patches;  % synthesis filters for patching
    
end

lpf257a = load_LPfilter;        % filters for low pass filtering of the input signal
llp = ((length(lpf257a)-1)/2);  % getting half the number of taps of the LP filter

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% defining the corresponding othonormal polynomials:
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

q = 1.1;  % bandwidth extension factor for the filterbank used for the noisy signals

cf = pi/2*q*(1:9)./sqrt(4*(1:9).^2-1);  % three term recursion coefficients for the polys

P = zeros(10,10);

P(1,:) = [zeros(1,9) 1];
P(2,:) = [zeros(1,8) 1/cf(1) 0];

for i=3:10
    P(i,:) = (1/cf(i-1)*[P(i-1,2:10) 0] - cf(i-2)/cf(i-1)*P(i-2,:));
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%
% GENERATING INPUT SIGNAL
%%%%%%%%%%%%%%%%%%%%%%%%%%%

time =(-nn:nn);  % vector of time indices, centered symmetrically around 0

s1 = A*(exp(dmp.'*time).*cos(f.'*time + phi.'*ones(1,NN)));    % generate a piece of the intended signal

if patching == 0
    
    if  win_type == 0
        win = ones(1,NN);
    end
    
    if  win_type == 1;
        win_cutoff = 8;   % select the cutoff for the ends of the widow -
        % must be an integer between 2 to 9
        
        win = CDwindow(win_cutoff,nn);
    end
    
    if  win_type == 2
        win = (kaiser(NN,3))';
    end
    
    if  win_type == 3
        win = (hamming(NN))';
    end
    
    wsig = s1 .* win;
    psig = [zeros(1,2*llp) wsig zeros(1,2*llp)];
    lh = length(psig);
    
else
    lsupport = (min(time)-lcd : min(time)+lcd);
    rsupport = (max(time)-lcd : max(time)+lcd);
    
    ls = A*cos(f.'*lsupport + phi.'*ones(1,2*lcd+1));
    rs = A*cos(f.'*rsupport + phi.'*ones(1,2*lcd+1));
    
    pchsig = [ls*CDClean*RPatch s1 rs*CDClean*LPatch];
    psig = [zeros(1,2*llp-31) pchsig zeros(1,2*llp-31)];
    lh = length(psig);
    psig = (psig/norm(psig))*sqrt(lh);
end





% adding noise

noise = randn(1,lh);

nsig = psig + ns*noise/norm(noise)*sqrt(lh);   %noisy signal

% low-pas filtering of the signal


sig1 = filter(lpf257a,1,nsig);          % filtering the signal to the bandwidth

sig = [sig1(llp+1:end) zeros(1,llp)];   % compensating for the group delay
% so that the filtered signal can be compared with
% the initial signal
% obtaining CDs of the signal

DF = zeros(9,lh);

for i=1:9
    DF(i,:)= filter(CD(i,:),1,sig);
end

dsign =[DF(:,lcd+1:end) zeros(9,lcd)];  % compensating for the filter delay

center = (lh-1)/2+1;   % center of the signal

time_test = -lcd-nn:lcd+nn;
dtest = A*(exp(dmp.'*time_test).*cos(f.'*time_test + phi.'*ones(1,length(time_test))));

test1 = zeros(9,length(time_test));
for i=1:9
    test1(i,:) = filter(CD(i,:),1,dtest);
end

test = test1(:,2*lcd+1:end);

adsign = zeros(10,NN); % generatig analytically computed derivatives of the input signal

for i=0:4
    adsign(2*i+1,:)= A*(-1)^floor((2*i+1)/2).*polyval(P(2*i+1,:),f)*cos(f.'*time + phi.'*ones(1,NN));
    adsign(2*i+2,:)= A*(-1)^floor(i+1).*polyval(P(2*i+2,:),f)* sin(f.'* time + phi.'*ones(1,NN));
end


sig_axis = 1:NN;
s_comp =cos(f.'*(1:NN) + phi.'*ones(1,NN));

ss = zeros(numsin,NN);
for i=1:numsin
    ss(i,:) = A(i)*s_comp(i,:);
end

figure
plot(sig_axis,ss);
hold on
plot(sig_axis,ns*noise(center-nn:center+nn),'c');
title('signal components and noise (cyan)');
figure
plot(sig_axis,sum(ss,1)+ns*noise(center-nn:center+nn),sig_axis,ns*noise(center-nn:center+nn),sig_axis,sum(ss,1));
title('signal, noise and noisy signal');

figure
plot(1:length(psig),psig);
hold on
plot(1:length(psig),ns*noise,'c');
title('signal components and noise (cyan)');
figure
plot(1:length(psig),psig,1:length(psig),ns*noise+psig);
title('signal, noise and noisy signal');
 
% computing products of CDs pointwise

prod = zeros(9,9,lh);


for k = 1:9
    for m = 1:k
        prod(k,m,:) = dsign(k,:).*conj(dsign(m,:));
        prod(m,k,:) = prod(k,m,:);
    end
end

% computing CD correlations:

cor = zeros(4,9,9,lh);
COR = zeros(4,9,9,lh);
Z = zeros(4,9,9,floor(interval(3)/2)+1);

for kk = 1:4
    for jj = 1:9
        for m = 1:jj
            cor(kk,jj,m,:) =(1/interval(kk))* filter(ones(1,interval(kk)),1,prod(jj,m,:));
            cor(kk,m,jj,:) = cor(kk,jj,m,:);
        end
    end
end


for kk =1:4
    cor(kk,:,:,:) = cat(4, cor(kk,:,:,floor(interval(kk)/2)+1:end), Z(kk,:,:,1:floor(interval(kk)/2)));
    COR(kk,1:2*kk+1,1:2*kk+1,:) = cor(kk,1:2*kk+1,1:2*kk+1,:);
end

if damped == 1
    
    Val = zeros(4,lh);
    Vec = zeros(4,9,lh);
    
    for i=1:lh
        for kk = 1:4
            DT = zeros(2*kk+1,2*kk+1);
            DT(:,:) = COR(kk,1:2*kk+1,1:2*kk+1,i);
            if abs(det(DT)) > 1e-17
                [Vec(kk,1:2*kk+1,i),Val(kk,i)] = eigs(DT,1,0);
                Vec(kk,:,i) = (ones(1,9)*sign(Vec(kk,2*kk+1,i))).*Vec(kk,:,i);
            end
        end
    end
    
%     VecM = zeros(4,9,lh);
%     for i = 1:4
%         VecM(i,:,:) = (1/aver_int)* filter(ones(1,aver_int),1,squeeze(Vec(i,:,:))')';
%         VecM(i,:,:) = cat(3,VecM(i,:,(aver_int-1)/2+1:end), zeros(1,9,(aver_int-1)/2));
%     end
%     
%     rumble = zeros(4,lh);
%     
%     for i=1:lh
%         for kk = 1:4
%             rumble(kk,i) = 1-(squeeze(VecM(kk,:,i))*squeeze(Vec(kk,:,i))')^2;
%         end
%     end
%     
%     norms = zeros(4,4);
%     for ii = 1:4
%         norms(ii) = norm(rumble(ii,center-K:center+K));
%     end
    
    % x_axis=(center-K:center+K);
    % figure
    % plot(x_axis,rumble(1,center-K:center+K),x_axis,rumble(2,center-K:center+K),x_axis,rumble(3,center-K:center+K))
    % legend(['total rumble 1 = ' num2str(norms(1))],['total rumble 2 = ' num2str(norms(2))],['total rumble 3 = ' num2str(norms(3))]);
    % title(['rumble, numsin =  ' num2str(numsin)]);
    
    VecF = median(Vec(numsin,:,center-K:center+K),3);
    
    PP = zeros(10,10);
    
    for i=1:10
        PP(i,:) = (1i)^(i-1)*(P(i,:).*(-1i).^(9:-1:0));
    end
    
    dmp_est=zeros(1,numsin);
    
    rts = roots(VecF*PP(1:9,:));
    freq_est1 = sort(imag(rts));
    freq_est = freq_est1(numsin+1:2*numsin)';
    for i=1:numsin
        dmp_est(i) = real(rts(imag(rts)==freq_est(i)));
    end
    err_freq = abs(f-freq_est);
    err_dmp = abs(dmp-dmp_est);
    
    
    
    disp(['true freq. and damp factors:    ' num2str(f) '        ' num2str(dmp)]);
    disp(['estimated freq and damp factors:    ' num2str(freq_est) '        ' num2str(dmp_est)]);
    disp(['estimation errors:   ' num2str(err_freq) '        ' num2str(err_dmp)]);
    
    out = VecF*dsign(1:end,:);
end

if damped == 0
    
    Val = zeros(4,lh);
    Vec = zeros(4,5,lh);
    
    for i=1:lh
        for kk = 1:4
            DT = zeros(kk+1,kk+1);
            DT(:,:) = COR(kk,1:2:2*kk+1,1:2:2*kk+1,i);
            if abs(det(DT)) > 1e-17
                [Vec(kk,1:kk+1,i),Val(kk,i)] = eigs(DT,1,0);
                Vec(kk,:,i) =(ones(1,5)*sign(Vec(kk,kk+1,i))).*Vec(kk,:,i);
            end
        end
    end
    
    
    
    % x_axis=(center-K:center+K);
    % figure
    % plot(x_axis,sqrt(Val(1,center-K:center+K)),x_axis,sqrt(Val(2,center-K:center+K)),x_axis,ns*ones(1,2*K+1));
    % title(['eigenvalues compared  ' num2str(ns)]);
    % legend('2-by-2 COR matrix','3-by-3 COR matrix');
    
    % figure
    % plot((0:lh-1),sqrt(Val(1,:)),'b')
    % hold on
    % plot((0:lh-1),sqrt(Val(2,:)),'r')
    % plot((0:lh-1),ns*ones(1,lh),'g');
    
    
    
    
%     VecM = zeros(4,5,lh);
%     for i = 1:4
%         VecM(i,:,:) = (1/aver_int)* filter(ones(1,aver_int),1,squeeze(Vec(i,:,:))')';
%         VecM(i,:,:) = cat(3,VecM(i,:,(aver_int-1)/2+1:end), zeros(1,5,(aver_int-1)/2));
%     end
%     
%     % figure
%     % plot((1:lh),squeeze(VecM(1,1,:)),(1:lh),squeeze(Vec(1,1,:)))
%     
%     rumble = zeros(4,lh);
%     
%     for i=1:lh
%         for kk = 1:4
%             rumble(kk,i) = 1-(squeeze(VecM(kk,:,i))*squeeze(Vec(kk,:,i))')^2;
%         end
%     end
%     
%     norms = zeros(4,4);
%     for ii = 1:4
%         norms(ii) = norm(rumble(ii,center-K:center+K));
%     end
    
    % x_axis=(center-K:center+K);
    % figure
    % plot(x_axis,rumble(1,center-K:center+K),x_axis,rumble(2,center-K:center+K),x_axis,rumble(3,center-K:center+K))
    % legend(['total rumble 1 = ' num2str(norms(1))],['total rumble 2 = ' num2str(norms(2))],['total rumble 3 = ' num2str(norms(3))]);
    % title(['rumble, numsin =  ' num2str(numsin)]);
    
    VecF = median(Vec(numsin,:,center-K:center+K),3);
    
    PP = zeros(10,10);
    
    for i=1:10
        PP(i,:) = (1i)^(i-1)*(P(i,:).*(-1i).^(9:-1:0));
    end
    
    dmp_est=zeros(1,numsin);
    
    rts = roots(VecF*PP(1:2:9,:));
    freq_est1 = sort(imag(rts));
    freq_est = freq_est1(numsin+1:2*numsin)';
    err_freq = abs(f-freq_est);
    
    
    disp(['true frequencies:    ' num2str(f)]);
    disp(['estimated frequencies:    ' num2str(freq_est)]);
    disp(['estimation errors:   ' num2str(err_freq)]);
    out = VecF*dsign(1:2:end,:);
    
end
nlevel = nsi*ones(1,lh);
figure
plot(1:lh,sqrt(Val),1:lh,nlevel)
title(['eigenvalues compared; noise =  ' num2str(ns/sqrt(2))]);

clean = psig-out;
figure
plot(1:lh,clean,1:lh,sig,1:lh,psig)
title('filtered signal');



fftlh = 8192*4;
FF = fft(sig,fftlh);
[a,b] = max(abs(FF));
PF = fft(psig,fftlh);
OUT = fft(out,fftlh);
figure
plot(1:fftlh/4,abs(PF(1:fftlh/4)),'r')
hold on
plot(1:fftlh/4,abs(FF(1:fftlh/4)),'b')
plot(1:fftlh/4,abs(OUT(1:fftlh/4)),'g')
stem(fftlh*f/(2* pi),1.1*max(abs(FF))*ones(1,numsin),'r');
stem(fftlh*freq_est/(2* pi),1.05*max(abs(FF))*ones(1,numsin),'g');
stem(b,1.15*a,'b');


figure
plot(1:fftlh/4,abs(PF(1:fftlh/4)),'r')
hold on
plot(1:fftlh/4,abs(FF(1:fftlh/4)),'b')
plot(1:fftlh/4,abs(OUT(1:fftlh/4)),'g')
stem(fftlh*f/(2* pi),1.1*max(abs(FF))*ones(1,numsin),'r');
stem(fftlh*freq_est/(2* pi),1.05*max(abs(FF))*ones(1,numsin),'g');
stem(b,1.15*a,'b');

% figure
% plot(1:4096/2,abs(OUT(1:4096/2)),'g')
% hold on
% stem(4096*f/(2* pi),1.1*max(abs(FF))*ones(1,numsin),'r');
% stem(4096*freq_est/(2* pi),1.05*max(abs(FF))*ones(1,numsin),'g');
% stem(b,1.15*a,'b');
% 

% NS = abs(fft(ns*noise,4096));
% figure
% plot(1:4096/4,NS(1:end/4),'r')
% hold on
% plot(1:4096/4,OUT(1:end/4),'b')
% stem(4096*f/(2* pi),1.1*max(abs(OUT))*ones(1,numsin),'r');
% stem(4096*freq_est/(2* pi),1.05*max(abs(OUT))*ones(1,numsin),'g');

